open Errors
open Abstract_syntax_tree

type 'a add_bottom = Nb of 'a | Bot

module type INT_DOMAIN =
sig
  module E: ERRORS
  module C: Constrain.CONSTRAIN
  type err = E.t
  type t
  val of_int: Z.t -> t
  val rand: Z.t -> Z.t -> t

  val pp: Format.formatter -> t -> unit

  val union: t -> t -> t
  val inter: t -> t -> t
  val widen: t -> t -> t
  val narrow: t -> t -> t

  val add: t -> t -> extent -> t add_bottom * err
  val sub: t -> t -> extent -> t add_bottom * err
  val times: t -> t -> extent -> t add_bottom * err
  val div: t -> t -> extent -> t add_bottom * err
  val modulo: t -> t -> extent -> t add_bottom * err

  val uplus: t -> extent -> t * err
  val uminus: t -> extent -> t * err

  val less: t -> t -> extent -> bool * bool (* may_be_true, may_be_false *)
  val less_eq: t -> t -> extent -> bool * bool
  val greater: t -> t -> extent -> bool * bool
  val greater_eq: t -> t -> extent -> bool * bool
  val eq: t -> t -> extent -> bool * bool
  val not_eq: t -> t -> extent -> bool * bool

  val get_constrains: t -> C.t
  val use_constrains: C.t -> t -> t
end

module MakeInterval (E: ERRORS) : INT_DOMAIN with module E = E and module C = Constrain.NonRelationalConstrain =
  (struct
    module E = E
    module C = Constrain.NonRelationalConstrain
    type err = E.t
    type t
    let of_int: Z.t -> t = fun x -> assert false
    let rand: Z.t -> Z.t -> t = fun a b -> assert false
    let add: t -> t -> extent -> t add_bottom * err = assert false
    let sub: t -> t -> extent -> t add_bottom * err = assert false
    let times: t -> t -> extent -> t add_bottom * err = assert false
    let div: t -> t -> extent -> t add_bottom * err = assert false
    let modulo: t -> t -> extent -> t add_bottom * err = assert false
    let uplus: t -> extent -> t * err = assert false
    let uminus: t -> extent -> t * err = assert false
    let less: t -> t -> extent -> bool * bool = assert false
    let less_eq: t -> t -> extent -> bool * bool = assert false
    let greater: t -> t -> extent -> bool * bool = assert false
    let greater_eq: t -> t -> extent -> bool * bool = assert false
    let eq: t -> t -> extent -> bool * bool = assert false
    let not_eq: t -> t -> extent -> bool * bool = assert false
    let pp: Format.formatter -> t -> unit = assert false
    let union: t -> t -> t = assert false
    let inter: t -> t -> t = assert false
    let widen: t -> t -> t = assert false
    let narrow: t -> t -> t = assert false
    let get_constrains: t -> C.t = assert false
    let use_constrains: C.t -> t -> t = assert false
  end)

module MakeParity (E: ERRORS) : INT_DOMAIN with module E = E and module C = Constrain.NonRelationalConstrain =
  (struct
    module E = E
    module C = Constrain.NonRelationalConstrain
    type err = E.t
    type t = Odd | Even | Top
    let is_empty: t -> bool = assert false
    let empty: t = assert false
    let of_int: Z.t -> t = assert false
    let rand: Z.t -> Z.t -> t = assert false
    let add: t -> t -> extent -> t add_bottom * err = assert false
    let sub: t -> t -> extent -> t add_bottom * err = assert false
    let times: t -> t -> extent -> t add_bottom * err = assert false
    let div: t -> t -> extent -> t add_bottom * err = assert false
    let modulo: t -> t -> extent -> t add_bottom * err = assert false
    let uplus: t -> extent -> t * err = assert false
    let uminus: t -> extent -> t * err = assert false
    let less: t -> t -> extent -> bool * bool = assert false
    let less_eq: t -> t -> extent -> bool * bool = assert false
    let greater: t -> t -> extent -> bool * bool = assert false
    let greater_eq: t -> t -> extent -> bool * bool = assert false
    let eq: t -> t -> extent -> bool * bool = assert false
    let not_eq: t -> t -> extent -> bool * bool = assert false
    let pp: Format.formatter -> t -> unit = assert false
    let union: t -> t -> t = assert false
    let inter: t -> t -> t = assert false
    let widen: t -> t -> t = assert false
    let narrow: t -> t -> t = assert false
    let get_constrains: t -> C.t = assert false
    let use_constrains: C.t -> t -> t = assert false
  end)

module MakeConcrete (E: ERRORS) : INT_DOMAIN with module E = E and module C = Constrain.NoConstrain =
  (struct
    module E = E
    module C = Constrain.NoConstrain
    type err = E.t
    type t
    let rec rand: Z.t -> Z.t -> t = fun a b -> assert false
    let of_int: Z.t -> t = fun x -> assert false
    let bin (f: Z.t -> Z.t -> Z.t) : t -> t -> extent -> t add_bottom * err = fun a b _ ->
      assert false
    let add: t -> t -> extent -> t add_bottom * err = bin Z.add
    let sub: t -> t -> extent -> t add_bottom * err = bin Z.sub
    let times: t -> t -> extent -> t add_bottom * err = bin Z.mul
    let div: t -> t -> extent -> t add_bottom * err = fun a b ext ->
      assert false
    let modulo: t -> t -> extent -> t add_bottom * err = fun a b ext ->
      assert false
    let uplus: t -> extent -> t * err = fun a _ -> assert false
    let uminus: t -> extent -> t * err = fun a _ -> assert false
    let bin_bool (f: Z.t -> Z.t -> bool): t -> t -> extent -> bool * bool = fun a b ext ->
      assert false
    let less: t -> t -> extent -> bool * bool = bin_bool Z.lt
    let less_eq: t -> t -> extent -> bool * bool = bin_bool Z.leq
    let greater: t -> t -> extent -> bool * bool = bin_bool Z.gt
    let greater_eq: t -> t -> extent -> bool * bool = bin_bool Z.geq
    let eq: t -> t -> extent -> bool * bool = bin_bool Z.equal
    let not_eq: t -> t -> extent -> bool * bool = bin_bool (fun a b -> Z.equal a b |> not)
    let pp: Format.formatter -> t -> unit = fun fmt t -> assert false
    let union (a: t) (b: t) : t = assert false
    let inter (a: t) (b: t) : t = assert false
    let widen: t -> t -> t = union
    let narrow: t -> t -> t = inter
    let get_constrains: t -> C.t = fun _ -> C.empty
    let use_constrains: C.t -> t -> t = fun _ a -> a
  end)
