open Abstract_syntax_tree
open Int_domain

module type VALUE_BASICS =
  (sig
    module E : Errors.ERRORS
    module C : Constrain.CONSTRAIN
    module I : Int_domain.INT_DOMAIN with module E = E and module C = C
    module B : Bool_domain.BOOL_DOMAIN with module E = E and module C = C
    type err = E.t
    type ival
    val compare: ival -> ival -> int
    val bot: ival
    val int_bool_binary_bool: (I.t -> I.t -> extent -> bool * bool) -> (B.t -> B.t -> bool) -> ival -> ival -> extent -> ival * err
    val int_unary: (I.t -> extent -> I.t * err) -> ival -> extent -> ival * err
    val int_binary: (I.t -> I.t -> extent -> I.t add_bottom * err) -> ival -> ival -> extent -> ival * err
    val int_binary_bool: (I.t -> I.t -> extent -> bool * bool) -> ival -> ival -> extent -> ival * err
    val bool_binary: (B.t -> B.t -> B.t) ->ival -> ival -> extent -> ival * err
    val bool_unary: (B.t -> B.t) -> ival -> extent -> ival * err
    val of_int: Z.t -> ival
    val of_bool: bool -> ival
    val rand: Z.t -> Z.t -> ival * err
    val pp_ival: Format.formatter -> ival -> unit
    val can_be_true: ival -> extent -> bool * err
    val can_be_false: ival -> extent -> bool * err

    val union: ival -> ival -> ival
    val inter: ival -> ival -> ival
    val widen: ival -> ival -> ival
    val narrow: ival -> ival -> ival
  end)

module type VALUE =
  (sig
    module E : Errors.ERRORS
    module C : Constrain.CONSTRAIN
    module I : Int_domain.INT_DOMAIN with module E = E and module C = C
    module B : Bool_domain.BOOL_DOMAIN with module E = E and module C = C
    type ival
    type err = E.t
    val compare: ival -> ival -> int
    val bot: ival
    val add: ival -> ival -> extent -> ival * err
    val sub: ival -> ival -> extent -> ival * err
    val times: ival -> ival -> extent -> ival * err
    val div: ival -> ival -> extent -> ival * err
    val modulo: ival -> ival -> extent -> ival * err

    val uplus: ival -> extent -> ival * err
    val uminus: ival -> extent -> ival * err

    val less: ival -> ival -> extent -> ival * err
    val less_eq: ival -> ival -> extent -> ival * err
    val greater: ival -> ival -> extent -> ival * err
    val greater_eq: ival -> ival -> extent -> ival * err

    val eq: ival -> ival -> extent -> ival * err
    val not_eq: ival -> ival -> extent -> ival * err

    val logical_or: ival -> ival -> extent -> ival * err
    val logical_and: ival -> ival -> extent -> ival * err
    val logical_not: ival -> extent -> ival * err

    val of_int: Z.t -> ival
    val of_bool: bool -> ival

    val pp_ival: Format.formatter -> ival -> unit

    val rand: Z.t -> Z.t -> ival * err

    val union: ival -> ival -> ival
    val inter: ival -> ival -> ival
    val widen: ival -> ival -> ival
    val narrow: ival -> ival -> ival
    val can_be_true: ival -> extent -> bool * err
    val can_be_false: ival -> extent -> bool * err

    val get_constrains: ival -> C.t
    val use_constrains: C.t -> ival -> ival
  end)

(* This should help you to write the next functor. *)
module MakeValueBasics(I: Int_domain.INT_DOMAIN)(B: Bool_domain.BOOL_DOMAIN with module C = I.C and module E = I.E) : VALUE_BASICS with module C = I.C and module E = I.E and module I = I and module B = B =
  (struct
    module E = I.E
    module B = B
    module I = I
    module C = I.C
    type err = E.t
    type ival = {int: I.t add_bottom; bool: B.t add_bottom}
    let compare a b =
      compare (compare a.int b.int) (compare a.bool b.bool)
    let bot = {int = Bot; bool = Bot}
    let int_bool_binary_bool: (I.t -> I.t -> extent -> bool * bool) -> (B.t -> B.t -> bool) -> ival -> ival -> extent -> ival * err =
      fun fi fb {int=ia; bool=ba} {int=ib; bool=bb} ext ->
      assert false

    let int_unary: (I.t -> extent -> I.t * err) -> ival -> extent -> ival * err =
      fun f a ext ->
      assert false

    let int_binary: (I.t -> I.t -> extent -> I.t add_bottom * err) -> ival -> ival -> extent -> ival * err =
      fun fi {int=ia; bool=ba} {int=ib; bool=bb} ext ->
      assert false

    let int_binary_bool: (I.t -> I.t -> extent -> bool * bool) -> ival -> ival -> extent -> ival * err =
      fun f {int=ia; bool=ba} {int=ib; bool=bb} ext ->
      assert false

    let bool_binary: (B.t -> B.t -> B.t) ->ival -> ival -> extent -> ival * err =
      fun fb {int=ia; bool=ba} {int=ib; bool=bb} ext ->
      assert false

    let bool_unary: (B.t -> B.t) -> ival -> extent -> ival * err =
      fun f a ext ->
      assert false

    let of_int: Z.t -> ival = fun n -> assert false
    let of_bool: bool -> ival = fun x -> assert false
    let rand: Z.t -> Z.t -> ival * err = fun x y -> assert false
    let can_be_true: ival -> extent -> bool * err =
      fun a ext ->
      assert false

    let can_be_false: ival -> extent -> bool * err =
      fun a ext ->
      assert false

    let pp_ival: Format.formatter -> ival -> unit =
      fun fmt a ->
      assert false

    let union: ival -> ival -> ival = fun a b ->
      assert false

    let inter: ival -> ival -> ival = fun a b ->
      assert false

    let widen: ival -> ival -> ival = fun a b ->
      assert false

    let narrow: ival -> ival -> ival = fun a b ->
      assert false
  end)

(* Don't forget to use VB *)
module MakeValue(VB: VALUE_BASICS) : VALUE with module C = VB.C and module E = VB.E and module I = VB.I and module B = VB.B and type ival = VB.ival =
  (struct
    module E = VB.E
    module I = VB.I
    module B = VB.B
    module C = VB.C
    module VB = VB
    type ival = VB.ival
    type err = VB.err
    let compare = VB.compare
    let bot = VB.bot
    let add: ival -> ival -> extent -> ival * err = VB.int_binary I.add
    let sub: ival -> ival -> extent -> ival * err = VB.int_binary I.sub
    let times: ival -> ival -> extent -> ival * err = VB.int_binary I.times
    let div: ival -> ival -> extent -> ival * err = VB.int_binary I.div
    let modulo: ival -> ival -> extent -> ival * err = VB.int_binary I.modulo
    let uplus: ival -> extent -> ival * err = VB.int_unary I.uplus
    let uminus: ival -> extent -> ival * err = VB.int_unary I.uminus
    let less: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.less
    let less_eq: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.less_eq
    let greater: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.greater
    let greater_eq: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.greater_eq
    let eq: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.eq
    let not_eq: ival -> ival -> extent -> ival * err = VB.int_binary_bool I.not_eq
    let logical_or: ival -> ival -> extent -> ival * err = VB.bool_binary B.logical_or
    let logical_and: ival -> ival -> extent -> ival * err = VB.bool_binary B.logical_and
    let logical_not: ival -> extent -> ival * err = VB.bool_unary B.logical_not
    let of_int: Z.t -> ival = VB.of_int
    let of_bool: bool -> ival = VB.of_bool
    let can_be_true: ival -> extent -> bool * err = VB.can_be_true
    let can_be_false: ival -> extent -> bool * err = VB.can_be_false
    let pp_ival: Format.formatter -> ival -> unit = VB.pp_ival
    let rand: Z.t -> Z.t -> ival * err = VB.rand
    let union: ival -> ival -> ival = VB.union
    let inter: ival -> ival -> ival = VB.inter
    let widen: ival -> ival -> ival = VB.widen
    let narrow: ival -> ival -> ival = VB.narrow

    let get_constrains: ival -> C.t = fun _ -> C.empty
    let use_constrains: C.t -> ival -> ival = fun _ a -> a
  end)

module Make(I: Int_domain.INT_DOMAIN)(B: Bool_domain.BOOL_DOMAIN with module C = I.C and module E = I.E) : VALUE with module C = I.C and module E = I.E and module I = I and module B = B = MakeValue(MakeValueBasics(I)(B))
